# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
simplified_test_function - Pytest extension for simplifying test functions.
"""

from __future__ import absolute_import

import re
import functools
from collections import namedtuple
try:
    from inspect import Signature, Parameter
except ImportError:  # py2
    from funcsigs import Signature, Parameter
import six
import pytest

__all__ = ['simplified_test_function']


# Pytest determines the signature of the test function by unpacking any
# wrapped functions (this is the default of the signature() function it
# uses. We correct this behavior by setting the __signature__ attribute of
# the wrapper function to its correct signature. To do that, we cannot use
# signature() because its follow_wrapped parameter was introduced only in
# Python 3.5. Instead, we build the signature manually.
TESTFUNC_SIGNATURE = Signature(
    parameters=[
        Parameter('desc', Parameter.POSITIONAL_OR_KEYWORD),
        Parameter('kwargs', Parameter.POSITIONAL_OR_KEYWORD),
        Parameter('exp_exc_types', Parameter.POSITIONAL_OR_KEYWORD),
        Parameter('exp_warn_types', Parameter.POSITIONAL_OR_KEYWORD),
        Parameter('condition', Parameter.POSITIONAL_OR_KEYWORD),
    ]
)


def simplified_test_function(test_func):
    """
    A decorator for test functions that simplifies the test function by
    handling a number of things:

    * Skipping the test if the `condition` item in the testcase is `False`,
    * Invoking the Python debugger if the `condition` item in the testcase is
      the string "pdb",
    * Capturing and validating any warnings issued by the test function,
      if the `exp_warn_types` item in the testcase is set,
    * Catching and validating any exceptions raised by the test function,
      if the `exp_exc_types` item in the testcase is set.

    This is a signature-changing decorator. This decorator must be inserted
    after the `pytest.mark.parametrize` decorator so that it is applied
    first (see the example).

    Parameters of the wrapper function returned by this decorator:

    * desc (string): Short testcase description.

    * kwargs (dict): Keyword arguments for the test function.

    * exp_exc_types: Expected exceptions, as one of the following:
        - exception type: This exception is expected.
        - tuple of exception types: Any of these exceptions is expected.
        - tuple of one exception type, one string: This exception is
          expected and its str() representation must match this regex pattern.
        - None: No exception is expected.

    * exp_warn_types (Warning or list of Warning): Expected warning types,
      or `None` if no warnings are expected.

    * condition (bool or 'pdb'): Boolean condition for running the testcase.
      If it evaluates to `bool(False)`, the testcase will be skipped.
      If it evaluates to `bool(True)`, the testcase will be run.
      The string value 'pdb' will cause the Python pdb debugger to be entered
      before calling the test function.

    Parameters of the test function that is decorated:

    * testcase (testcase_tuple): The testcase, as a named tuple.

    * **kwargs: Keyword arguments for the test function.

    Example::

        TESTCASES_FOO_EQUAL = [
            # desc, kwargs, exp_exc_types, exp_warn_types, condition
            (
                "Equality with different lexical case of name",
                dict(
                    obj1=Foo('Bar'),
                    obj2=Foo('bar'),
                    exp_equal=True,
                ),
                None, None, True
            ),
            # ... more testcases
        ]

        @pytest.mark.parametrize(
            "desc, kwargs, exp_exc_types, exp_warn_types, condition",
            TESTCASES_FOO_EQUAL)
        @pytest_extensions.simplified_test_function
        def test_Foo_equal(testcase, obj1, obj2, exp_equal):

            # The code to be tested
            equal = (obj1 == obj2)

            # Ensure that exceptions raised in the remainder of this function
            # are not mistaken as expected exceptions
            assert testcase.exp_exc_types is None

            # Verify the result
            assert equal == exp_equal
    """

    # A testcase tuple
    testcase_tuple = namedtuple(
        'testcase_tuple',
        ['desc', 'kwargs', 'exp_exc_types', 'exp_warn_types', 'condition']
    )

    def wrapper_func(desc, kwargs, exp_exc_types, exp_warn_types, condition):
        """
        Wrapper function that calls the test function that is decorated.
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        if condition == 'pdb':
            # pylint: disable=import-outside-toplevel
            import pdb

        testcase = testcase_tuple(desc, kwargs, exp_exc_types, exp_warn_types,
                                  condition)

        # Process the exp_exc_types parameter and pull out the message pattern
        # from the list.
        exp_exc_pattern = None  # Expected exception message regexp pattern
        if isinstance(exp_exc_types, (list, tuple)):
            if len(exp_exc_types) == 2 and \
                    issubclass(exp_exc_types[0], Exception) and \
                    isinstance(exp_exc_types[1], six.string_types):
                exp_exc_pattern = exp_exc_types[1]
                exp_exc_types = exp_exc_types[0]

        if exp_warn_types:
            with pytest.warns(exp_warn_types) as rec_warnings:
                if exp_exc_types:
                    with pytest.raises(exp_exc_types) as exc_info:
                        if condition == 'pdb':
                            pdb.set_trace()

                        test_func(testcase, **kwargs)  # expecting an exception

                    ret = None  # Test function has returned (debugging hint)

                    # Verify the exception message pattern, if specified
                    if exp_exc_pattern:
                        exc_message = str(exc_info.value)
                        m = re.search(exp_exc_pattern, exc_message)
                        assert m, \
                            "Unexpected exception message:\n" \
                            "  Expected pattern: {exp}\n" \
                            "  Actual message: {act}\n". \
                            format(act=exc_message, exp=exp_exc_pattern)

                    # In combination with exceptions, we do not verify warnings
                    # (they could have been issued before or after the
                    # exception).
                else:
                    if condition == 'pdb':
                        pdb.set_trace()

                    test_func(testcase, **kwargs)  # not expecting an exception

                    ret = None  # Test function has returned (debugging hint)

                    assert len(rec_warnings) >= 1
        else:
            with pytest.warns(None) as rec_warnings:
                if exp_exc_types:
                    with pytest.raises(exp_exc_types) as exc_info:
                        if condition == 'pdb':
                            pdb.set_trace()

                        test_func(testcase, **kwargs)  # expecting an exception

                    ret = None  # Test function has returned (debugging hint)

                    # Verify the exception message pattern, if specified
                    if exp_exc_pattern:
                        exc_message = str(exc_info.value)
                        m = re.search(exp_exc_pattern, exc_message)
                        assert m, \
                            "Unexpected exception message:\n" \
                            "  Expected pattern: {exp}\n" \
                            "  Actual message: {act}\n". \
                            format(act=exc_message, exp=exp_exc_pattern)

                else:
                    if condition == 'pdb':
                        pdb.set_trace()

                    test_func(testcase, **kwargs)  # not expecting an exception

                    ret = None  # Test function has returned (debugging hint)

                    # Verify that no warnings have occurred
                    if exp_warn_types is None and rec_warnings:
                        lines = []
                        for w in rec_warnings.list:
                            tup = (w.filename, w.lineno, w.category.__name__,
                                   str(w.message))
                            line = "{t[0]}:{t[1]}: {t[2]}: {t[3]}".format(t=tup)
                            if line not in lines:
                                lines.append(line)
                        msg = "Unexpected warnings:\n{}".format(
                            '\n'.join(lines))
                        raise AssertionError(msg)
        return ret

    # Needed because the decorator is signature-changin
    wrapper_func.__signature__ = TESTFUNC_SIGNATURE

    return functools.update_wrapper(wrapper_func, test_func)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The EasyVault class and its exceptions.
"""

from __future__ import absolute_import, print_function
import os
import re
import tempfile
import base64

import yaml
import six
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

__all__ = ['EasyVault', 'EasyVaultException', 'EasyVaultFileError',
           'EasyVaultDecryptError', 'EasyVaultEncryptError',
           'EasyVaultYamlError']


# Line width in encrypted vault file
ENCRYPTED_LINE_WIDTH = 80

# Header used in encrypted files
HEADER_MONIKER = b'EASY_VAULT'
HEADER_VERSION = b'1.0'
HEADER = b'$' + HEADER_MONIKER + b';' + HEADER_VERSION + b'\n'
# Python 2.7 does not support using both 'r' and 'b', so we use 'b' only.
HEADER_PATTERN = re.compile(
    b'^\\$(?P<moniker>[A-Z0-9_]+);(?P<version>[0-9.]+)$')


class EasyVaultException(Exception):
    """
    Base exception for all exceptions raised by the
    :class:`~easy_vault.EasyVault` class.

    Derived from :exc:`py:Exception`.
    """
    pass


class EasyVaultFileError(EasyVaultException):
    """
    Exception indicating file I/O errors with a vault file.

    Derived from :exc:`EasyVaultException`.
    """
    pass


class EasyVaultDecryptError(EasyVaultException):
    """
    Exception indicating that an encrypted vault file could not be decrypted.

    Derived from :exc:`EasyVaultException`.
    """
    pass


class EasyVaultEncryptError(EasyVaultException):
    """
    Exception indicating that an unencrypted vault file could not be encrypted.

    Derived from :exc:`EasyVaultException`.
    """
    pass


class EasyVaultYamlError(EasyVaultException):
    """
    Exception indicating that a vault file in YAML format has a format issue.

    Derived from :exc:`EasyVaultException`.
    """
    pass


class EasyVault(object):
    """
    A vault file that can be encrypted and decrypted.

    An object of this class is tied to a single vault file and a single
    vault password.

    There are no requirements for the format of the vault file. It may be a text
    file or a binary file (but the typical case for a vault file would be a text
    file, e.g. in YAML format).

    There is no size limit to the vault file. However, because the complete
    vault file is read into memory in one chunk, this implementation is not
    well suited for handling huge files. It is really meant for vault files:
    Files that keep secrets, but not huge data.

    The password is converted to a symmetric key which is then used for
    encryption and decryption of the vault file.

    The encryption package used by this class is pluggable. The default
    implementation uses the symmetric key support from the
    `cryptography package <https://cryptography.io/en/stable/>`_.

    Users who whish to use a different encryption package can do so by
    subclassing this class and implementing the following methods to use a
    different encryption package:

    * :meth:`generate_key` - Calculate a symmetric key from a password.
    * :meth:`encrypt_data` - Encrypt clear data with a symmetric key.
    * :meth:`decrypt_data` - Decrypt encrypted data with a symmetric key.
    """

    def __init__(self, filepath, password):
        """
        Parameters:

          filepath (:term:`unicode string`):
            Path name of the vault file.

          password (:term:`unicode string`):
            Password for encrypting and decrypting the vault file.
        """
        self._filepath = filepath
        if not isinstance(password, six.string_types):
            raise TypeError(
                "The password argument is not a unicode string, but: {t}".
                format(t=type(password)))
        key = self.generate_key(password)
        if not isinstance(key, six.binary_type):
            raise TypeError(
                "The return value of generate_key() is not a byte string, "
                "but: {t}".
                format(t=type(key)))
        self._key = key

    @property
    def filepath(self):
        """
        :term:`unicode string`): Path name of the vault file.
        """
        return self._filepath

    def is_encrypted(self):
        """
        Test whether the vault file is encrypted by easy-vault.

        This is done by checking for the unique easy-vault header in the first
        line of the vault file.

        Returns:
          bool: Boolean indicating whether the vault file is encrypted by
          easy-vault.

        Raises:
          EasyVaultFileError: Error with the vault file
        """

        try:
            with open(self._filepath, 'rb') as fp:
                first_bline = fp.readline()  # Including trailing newline
        except (OSError, IOError) as exc:
            new_exc = EasyVaultFileError(
                "Cannot open vault file {fn} for reading: {exc}".
                format(fn=self._filepath, exc=exc))
            new_exc.__cause__ = None
            raise new_exc  # VaultFileOpenError

        m = HEADER_PATTERN.match(first_bline)
        if m is None:
            return False
        if m.group('moniker') != HEADER_MONIKER:
            return False
        return True

    def encrypt(self):
        """
        Encrypt the vault file.

        The vault file must be unencrypted (i.e. not encrypted by easy-vault).

        Raises:
          EasyVaultFileError: Error with the vault file or a temporary file
          EasyVaultEncryptError: Error encrypting the vault file
        """

        try:
            with open(self._filepath, 'rb', encoding=None) as fp:
                first_bline = fp.readline()  # Including trailing newline
                m = HEADER_PATTERN.match(first_bline)
                if m and m.group('moniker') == HEADER_MONIKER:
                    raise EasyVaultEncryptError(
                        "Cannot encrypt vault file {fn}: "
                        "The file is already encrypted".
                        format(fn=self._filepath))
                clear_bdata = first_bline + fp.read()
        except (OSError, IOError) as exc:
            new_exc = EasyVaultFileError(
                "Cannot open vault file {fn} for reading: {exc}".
                format(fn=self._filepath, exc=exc))
            new_exc.__cause__ = None
            raise new_exc  # EasyVaultFileError

        assert isinstance(clear_bdata, six.binary_type), type(clear_bdata)
        encrypted_bdata = self.encrypt_data(clear_bdata, self._key)
        if not isinstance(encrypted_bdata, six.binary_type):
            raise TypeError(
                "The return value of encrypt_data() is not a byte string, "
                "but: {t}".format(t=type(encrypted_bdata)))

        encrypted_blines = HEADER
        for bchunk in chunks(encrypted_bdata, ENCRYPTED_LINE_WIDTH):
            encrypted_blines += bchunk + b'\n'

        write_file(self._filepath, encrypted_blines)

    def decrypt(self):
        """
        Decrypt the vault file.

        The vault file must be encrypted by easy-vault.

        Raises:
          EasyVaultFileError: Error with the vault file or a temporary file
          EasyVaultDecryptError: Error decrypting the vault file
        """
        clear_bdata = self._get_bytes_from_encrypted()
        write_file(self._filepath, clear_bdata)

    def _get_bytes_from_encrypted(self):
        """
        Get the unencrypted data from an encrypted vault file.

        Returns:
          :term:`byte string`: Unencrypted data from the vault file.

        Raises:
          EasyVaultFileError: Error with the vault file
          EasyVaultDecryptError: Error decrypting the vault file
        """

        try:
            with open(self._filepath, 'rb') as fp:
                first_bline = fp.readline()  # Including trailing newline
                m = HEADER_PATTERN.match(first_bline)
                if m is None or m.group('moniker') != HEADER_MONIKER:
                    raise EasyVaultDecryptError(
                        "Cannot decrypt vault file {fn}: "
                        "The file is not encrypted".
                        format(fn=self._filepath))
                encrypted_blines = fp.read()
        except (OSError, IOError) as exc:
            new_exc = EasyVaultFileError(
                "Cannot open vault file {fn} for reading: {exc}".
                format(fn=self._filepath, exc=exc))
            new_exc.__cause__ = None
            raise new_exc  # EasyVaultFileError

        encrypted_bdata = b'\n'.join(encrypted_blines.split(b'\n'))
        assert isinstance(encrypted_bdata, six.binary_type), \
            type(encrypted_bdata)
        clear_bdata = self.decrypt_data(encrypted_bdata, self._key)
        if not isinstance(clear_bdata, six.binary_type):
            raise TypeError(
                "The return value of decrypt_data() is not a byte string, "
                "but: {t}".format(t=type(clear_bdata)))
        return clear_bdata

    def _get_bytes_from_clear(self):
        """
        Get the data from an unencrypted vault file.

        Returns:
          :term:`byte string`: Unencrypted data from the vault file.

        Raises:
          EasyVaultFileError: Error with the vault file
        """
        try:
            with open(self._filepath, 'rb') as fp:
                clear_bdata = fp.read()
        except (OSError, IOError) as exc:
            new_exc = EasyVaultFileError(
                "Cannot open vault file {fn} for reading: {exc}".
                format(fn=self._filepath, exc=exc))
            new_exc.__cause__ = None
            raise new_exc  # EasyVaultFileError
        return clear_bdata

    def get_bytes(self):
        """
        Get the unencrypted content of the vault file as a Byte sequence.

        The vault file may be encrypted or unencrypted.

        Returns:
          :term:`byte string`: Unencrypted content of the vault file, as a Byte
          sequence.

        Raises:
          EasyVaultFileError: Error with the vault file
          EasyVaultDecryptError: Error decrypting the vault file
        """
        if self.is_encrypted():
            return self._get_bytes_from_encrypted()
        return self._get_bytes_from_clear()

    def get_text(self):
        """
        Get the unencrypted content of the vault file as a Unicode string.

        The vault file may be encrypted or unencrypted.

        Returns:
          :term:`unicode string`: Unencrypted content of the vault file, as a
          Unicode string.

        Raises:
          EasyVaultFileError: Error with the vault file
          EasyVaultDecryptError: Error decrypting the vault file
        """
        bdata = self.get_bytes()
        udata = bdata.decode('utf-8')
        return udata

    def get_yaml(self):
        """
        Get the unencrypted content of a YAML-formatted vault file as a YAML
        object.

        The vault file may be encrypted or unencrypted.

        Returns:
          dict or list: Top-level object of the YAML-formatted vault file.

        Raises:
          EasyVaultFileError: Error with the vault file or a temporary file
          EasyVaultYamlError: YAML syntax error in the vault file
          EasyVaultDecryptError: Error decrypting the vault file
        """
        clear_bdata = self.get_bytes()
        try:
            data_obj = yaml.safe_load(clear_bdata)
        except yaml.YAMLError as exc:
            new_exc = EasyVaultYamlError(
                "Invalid YAML syntax in vault file {fn}: {exc}".
                format(fn=self._filepath, exc=exc))
            new_exc.__cause__ = None
            raise new_exc  # EasyVaultYamlError
        return data_obj

    @staticmethod
    def generate_key(password):
        """
        Encryption implementation: Calculate a symmetric key from a password.

        The key must match the requirements of the encryption package that is
        used in the :meth:`encrypt_data` and :meth:`decrypt_data` methods.

        Using this method repeatedly on the same password must result in the
        same key.

        This method can be overwritten by users to use a different encryption
        package. Its default implementation uses the
        `cryptography package <https://cryptography.io/en/stable/>`_,
        and calculates the key as a 256-bit key using 10000 iterations of
        SHA256 on the password, using a fixed salt.

        Parameters:

          password (:term:`unicode string`):
            Password for encrypting and decrypting the vault file.

        Returns:
          :term:`byte string`: The calculated key.
        """
        salt = b'fixed'
        bpassword = password.encode('utf-8')
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000,
            backend=default_backend())
        key = base64.urlsafe_b64encode(kdf.derive(bpassword))
        return key

    @staticmethod
    def encrypt_data(clear_data, key):
        """
        Encryption implementation: Encrypt clear data with a symmetric key.

        This method can be overwritten by users to use a different encryption
        package. Its default implementation uses the :mod:`cryptography.fernet`
        encryption package.

        Parameters:

          clear_data (:term:`byte string`):
            The clear data to be encrypted.

          key (:term:`byte string`):
            The symmetric key to be used for the encryption.

        Returns:
          :term:`byte string`: The encrypted data.
        """
        f = Fernet(key)
        encrypted_data = f.encrypt(clear_data)
        return encrypted_data

    def decrypt_data(self, encrypted_data, key):
        """
        Encryption implementation: Decrypt encrypted data with a symmetric key.

        This method can be overwritten by users to use a different encryption
        package. Its default implementation uses the :mod:`cryptography.fernet`
        encryption package.

        Parameters:

          encrypted_data (:term:`byte string`):
            The encrypted data to be decrypted.

          key (:term:`byte string`):
            The encryption key to be used.

        Returns:
          :term:`byte string`: The clear data.

        Raises:
          EasyVaultDecryptError: Error decrypting the vault file
        """
        f = Fernet(key)
        try:
            clear_data = f.decrypt(encrypted_data)
        except InvalidToken:
            new_exc = EasyVaultDecryptError(
                "Cannot decrypt vault file {fn}: Invalid password".
                format(fn=self._filepath))
            new_exc.__cause__ = None
            raise new_exc  # EasyVaultDecryptError
        return clear_data


def write_file(filepath, data):
    """
    Write data to a file in a manner that is safe against keyboard breaks or
    other program aborts.

    This is done by first writing the data to a temporary file and then renaming
    the file. Since the rename operation is performed by the operating system
    as an atomic operation, it results either in the new file or the previous
    file still exists unchanged.

    Parameters:

      filepath (:term:`unicode string`):
        Path name of the file to be written.

      data (:term:`byte string`):
        The data to be written.

    Raises:
      EasyVaultFileError: Error with the vault file or a temporary file
    """
    assert isinstance(data, six.binary_type), type(data)
    try:
        with tempfile.NamedTemporaryFile(
                mode='wb', delete=False, prefix='easy_vault') as tfp:
            tfp.write(data)
            tfpath = tfp.name
    except (OSError, IOError) as exc:
        new_exc = EasyVaultFileError(
            "Cannot open temporary file {tfn} for writing: {exc}".
            format(tfn=tfp.name, exc=exc))
        new_exc.__cause__ = None
        raise new_exc  # EasyVaultFileError

    try:
        os.rename(tfpath, filepath)
    except (OSError, IOError) as exc:
        new_exc = EasyVaultFileError(
            "Cannot rename temporary file {tfn} to vault file {fn}: {exc}".
            format(tfn=tfpath, fn=filepath, exc=exc))
        new_exc.__cause__ = None
        raise new_exc  # EasyVaultFileError


def chunks(iterable, size):
    """
    Generator that yields successive equal-sized chunks from a subscriptable
    iterable (such as a string).
    """
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Convenience functions for getting and setting the password in the keyring.
"""

from __future__ import absolute_import, print_function

import getpass
from ._key_ring_lib import KeyRingLib

__all__ = ['get_password', 'set_password']


def get_password(filepath, use_keyring=True, verbose=False, echo=print):
    """
    Get the vault password from the keyring or by prompting for it.

    This is a convenience function that uses the :class:`~easy_vault.KeyRingLib`
    class.

    In non-keyring mode, the keyring is not used, and the password is always
    prompted for.

    In keyring mode, the password is attempted to be obtained from the keyring
    and is prompted for if not available there.

    Parameters:

      filepath (:term:`unicode string`):
        Path name of the vault file.

      use_keyring (bool):
        Use keyring mode (`True`) or non-keyring mode (`False`).

      verbose (bool):
        Print additional messages. Note that the password prompt is always
        displayed.

      echo (function):
        Print function to be used for the additional messages in verbose mode.

    Returns:
      :term:`unicode string`: Password for the vault file.
    """
    keyringlib = KeyRingLib()

    if not use_keyring:
        return getpass.getpass("Enter password for vault file {fn}:".
                               format(fn=filepath))

    password = keyringlib.get_password(filepath)
    if password is None:
        return getpass.getpass("Enter password for vault file {fn}:".
                               format(fn=filepath))

    if verbose:
        echo("Using password from keyring for vault file: {fn}".
             format(fn=filepath))
    return password


def set_password(
        filepath, password, use_keyring=True, verbose=False, echo=print):
    """
    Set the password in the keyring.

    This is a convenience function that uses the :class:`~easy_vault.KeyRingLib`
    class.

    In non-keyring mode, this function does nothing.

    In keyring mode, the password is stored in the keyring. This is done by
    first getting it, and setting it only if it was not set or was different.
    This approach has been chosen in order to print the verbose message
    about setting the password only if it was really changed.

    Parameters:

      filepath (:term:`unicode string`):
        Path name of the vault file.

      password (:term:`unicode string`):
        Password for the vault file.

      use_keyring (bool):
        Use keyring mode (`True`) or non-keyring mode (`False`).

      verbose (bool):
        Print additional messages.

      echo (function):
        Print function to be used for the additional messages in verbose mode.
    """
    if use_keyring:
        keyringlib = KeyRingLib()
        current_password = keyringlib.get_password(filepath)
        if current_password is None or password != current_password:
            if verbose:
                echo("Setting password in keyring for vault file: {fn}".
                     format(fn=filepath))
            keyringlib.set_password(filepath, password)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The KeyRingLib class.
"""

from __future__ import absolute_import, print_function
import os
import keyring

__all__ = ['KeyRingLib']


# Data stored in keyring enzry, by version:
#   1.0: The item stored is the password
KEYRING_SERVICE = u'easy_vault/pypi.org/1.0'    # Must be unique within keyring


class KeyRingLib(object):
    """
    Access to the local system keyring for storing easy-vault vault passwords.

    An object of this class is tied to the local keyring facility and can
    store and retrieve multiple vault passwords, one for each vault file.

    The keyring items that are created have a service name that starts with
    'easy_vault'. There is one keyring item for each vault file.
    """

    def get_password(self, filepath):
        """
        Get the password for the specified vault file from the keyring.

        If the keyring does not store a password for this vault file, `None`
        is returned.

        Parameters:

          filepath (:term:`unicode string`):
            Path name of the vault file. It will be normalized to identify the
            keyring entry for the vault file.

        Returns:
          :term:`unicode string`: Password for the vault file, or `None`.
        """
        return keyring.get_password(
            self.keyring_service(), self.keyring_username(filepath))

    def set_password(self, filepath, password):
        """
        Set the password for the specified vault file in the keyring.

        Parameters:

          filepath (:term:`unicode string`):
            Path name of the vault file. It will be normalized to identify the
            keyring entry for the vault file.

          password (:term:`unicode string`):
            Password for the vault file.
        """
        keyring.set_password(
            self.keyring_service(), self.keyring_username(filepath), password)

    @staticmethod
    def keyring_service():
        """
        Return the service name that is used for the :mod:`keyring` module.

        That name is fixed within easy-vault and starts with 'easy_vault'.

        Returns:
          :term:`unicode string`: keyring service name.
        """
        return KEYRING_SERVICE

    @staticmethod
    def keyring_username(filepath):
        """
        Return the user name that is used for the :mod:`keyring` module.

        That name is calculated from the normalized vault file path, such that
        each different vault file uses a different entry in the keyring.

        Parameters:

          filepath (:term:`unicode string`):
            Path name of the vault file.

        Returns:
          :term:`unicode string`: keyring user name.
        """
        normpath = os.path.abspath(filepath)
        return u'file:{fn}'.format(fn=normpath)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the _easy_vault.py module.
"""

from __future__ import absolute_import, print_function
import pytest
# from testfixtures import TempDirectory
from easy_vault import EasyVault

from ..utils.simplified_test_function import simplified_test_function


TEST_VAULT_FILEPATH = 'examples/vault.yml'
TEST_VAULT_PASSWORD = 'mypassword'

TESTCASES_VAULT_INIT = [

    # Testcases for EasyVault.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to EasyVault().
    #   * init_kwargs: Dict of keyword arguments to EasyVault().
    #   * exp_attrs: Dict with expected EasyVault attributes.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Order of positional parameters",
        dict(
            init_args=(
                TEST_VAULT_FILEPATH,
                TEST_VAULT_PASSWORD,
            ),
            init_kwargs=dict(),
            exp_attrs={
                'filepath': TEST_VAULT_FILEPATH,
            },
        ),
        None, None, True
    ),
    (
        "Names of keyword arguments",
        dict(
            init_args=(),
            init_kwargs=dict(
                filepath=TEST_VAULT_FILEPATH,
                password=TEST_VAULT_PASSWORD,
            ),
            exp_attrs={
                'filepath': TEST_VAULT_FILEPATH,
            },
        ),
        None, None, True
    ),
    (
        "Omitted required parameter: filepath",
        dict(
            init_args=(),
            init_kwargs=dict(
                password=TEST_VAULT_PASSWORD,
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Omitted required parameter: password",
        dict(
            init_args=(),
            init_kwargs=dict(
                filepath=TEST_VAULT_FILEPATH,
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Vault file does not exist (no error at init time)",
        dict(
            init_args=(),
            init_kwargs=dict(
                filepath='invalid_file',
                password=TEST_VAULT_PASSWORD,
            ),
            exp_attrs={
                'filepath': 'invalid_file',
            },
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_VAULT_INIT)
@simplified_test_function
def test_vault_init(testcase, init_args, init_kwargs, exp_attrs):
    """
    Test function for EasyVault.__init__()
    """

    # The code to be tested
    act_obj = EasyVault(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None, \
        "Expected exception not raised: {}". \
        format(testcase.exp_exc_types)

    for attr_name in exp_attrs:
        exp_attr_value = exp_attrs[attr_name]
        assert hasattr(act_obj, attr_name), \
            "Missing attribute {0!r} in returned EasyVault object". \
            format(attr_name)
        act_attr_value = getattr(act_obj, attr_name)
        assert act_attr_value == exp_attr_value, \
            "Unexpected value for attribute {0!r}: Expected {1!r}, got {2!r}".\
            format(attr_name, exp_attr_value, act_attr_value)
